package contrasts;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Map;
import JSci.maths.statistics.ChiSqrDistribution;
import JSci.maths.statistics.NormalDistribution;

/**
 * This is a static implementation of all implemented statistical methods on
 * contrasts.
 * 
 * This class provides the statistical methods of confidence intervals of
 * contrasts among multinomial populations. Three different methods are
 * implemented:
 * 
 * <ul>
 * <li>GOODMAN - Method of Goodman (see paper)</li>
 * <li>BONFERRONI - Simply Bonferroni aproach (BE CAREFULL: HIGH NUMBER OF FALSE
 * POSITIVES!)</li>
 * <li>GOODMAN+BONFERRONI - Uses only Bonferroni, if the method results in
 * smaller intervals. But this is nearly always the case because the maximum
 * number of used contrasts is extremely lower then the total amount.</li>
 * <li>HOLM-BONFERRONI - Creates the interval over the normal unity distribution
 * with confidence of alpha.</li>
 * </ul>
 * 
 * @author <a href="mailto:max.schubach@informatik.uni-tuebingen.de">Max
 *         Schubach</a>
 * 
 * @version 1.1
 * 
 */
public class Method {

	protected ArrayList<String> methods = new ArrayList<String>(
			Arrays.asList(new String[] { "GOODMAN", "BONFERRONI",
					"GOODMAN+BONFERRONI", "HOLM-BONFERRONI" }));
	private String method = "GOODMAN";
	/**
	 * The alpha of the statistical test (default is 0.05)
	 */
	private double alpha = 0.05;
	private int degreeOfFreedom = 0;
	private int numberOfContrasts = 0;
	private double pValueBorder;
	private int comparisons = 0;

	/**
	 * 
	 */
	public Method() {
		this.alpha = 0.05;
		this.method = "GOODMAN";
		this.degreeOfFreedom = 0;
		this.numberOfContrasts = 0;
	}

	/**
	 * @param method
	 */
	protected void setMethod(String method) {
		this.method = method;
		if (!this.methods.contains(method)) {
			System.err.println("CONTRASTS: Unknown metod \"" + this.method
					+ "\" - Using \"GOODMAN\" instead!");
			this.method = "GOODMAN";
		}

	}

	/**
	 * Sets different Method constants. The degree of freedom the number of
	 * comparisons and the number of contrasts. all of them can be computed by
	 * #populations and #classes
	 * 
	 * @param i
	 *            Number of populations
	 * @param j
	 *            Number of classes
	 */
	protected void setConstants(int i, int j) {
		this.degreeOfFreedom = (i - 1) * (j - 1);
		this.comparisons = (i * (i - 1)) / 2;
		this.numberOfContrasts = 2 * this.comparisons * j;
	}

	/**
	 * Setter method of alpha.
	 * 
	 * @param alpha
	 *            the wanted alpha
	 */
	protected void setAlpha(double alpha) {
		this.alpha = alpha;
	}

	/**
	 * Computes the inverse probability (percentile) of of the IxJ contingency
	 * table on the used distribution.
	 * 
	 * @param i
	 *            number of populations
	 * @param j
	 *            number of classes
	 * @return the inverse probability of the used distribution
	 */
	protected double getPercentile(int i, int j) {
		if (this.degreeOfFreedom == 0 || this.numberOfContrasts == 0
				|| this.comparisons == 0) {
			setConstants(i, j);
		}
		// System.out.println(Method.degreeOfFreedom);
		// System.out.println(j);

		double percentile;

		if (this.method.equals("GOODMAN")
				|| this.method.equals("GOODMAN+BONFERRONI")) {
			percentile = Math.sqrt(this.qchisq());
			// System.out.println(percentile);

			if (this.method.equals("GOODMAN+BONFERRONI")) {

				double z = computeBonferroniQNorm(i, j);

				if (z <= percentile) {
					percentile = z;
				}
			}

		} else if (this.method.equals("BONFERRONI")) {

			double z = computeBonferroniQNorm(i, j);
			percentile = z;

		} else if (this.method.equals("HOLM-BONFERRONI")) {
			percentile = qnorm(this.alpha);
		} else {
			System.err.println("Unknown method! Using \"GOODMAN\"");
			double l = Math.sqrt(this.qchisq());
			percentile = l;
		}

		return percentile;
	}

	private double computeBonferroniQNorm(int i, int j) {
		double z = this.qnorm(this.alpha / (2 * this.numberOfContrasts));
		return z;
	}

	/**
	 * Computes the 100(1-alpha) quartile of the chi-square distribution at a
	 * given degree of Freedom.
	 * 
	 * @param alpha
	 *            The needed alpha value.
	 * @param degreeOfFreedom
	 *            The degree of freedom of the chisquare distribution.
	 * @return 100(1-alpha) quartile of the chi-square distribution If
	 *         qchisq(alpha, df) * variance the confidence interval with 1-alpha
	 *         confidence level.
	 */
	private double qchisq() {

		ChiSqrDistribution chi = new ChiSqrDistribution(this.degreeOfFreedom);
		return chi.inverse(1 - this.alpha);

		// double[] chisquarePara = { Method.degreeOfFreedom };
		// double[] chisquarePercentiles = { 1.0, 2.0 };
		// double chisquareX = 1.0;
		// double chisquareProb = 1.0 - Method.alpha;
		// Probability chisquare = new Probability(DistributionType.CHISQUARE,
		// chisquarePara, chisquarePercentiles, chisquareX, chisquareProb);
		// return chisquare.inverse;
	}

	/**
	 * Quartile function of the unity normal distribution.
	 * 
	 * @param alpha
	 *            The needed alpha value.
	 * @return The quartile from alpha of the unity normal distribution. If
	 *         qnorm(alpha) * variance the confidence interval with 1-alpha
	 *         confidence level.
	 */
	@SuppressWarnings("static-method")
	private double qnorm(double alpha) {
		NormalDistribution normal = new NormalDistribution(0.0, Math.sqrt(1.0));
		return normal.inverse(1 - alpha);
		// double[] normalPara = { 0.0, 1.0 };
		// double[] normalPercentiles = { 1 - 1.96 * 2, 1 + 1.96 * 2 };
		// double normalX = normalPara[0];
		// double normalProb = 1.0 - alpha;
		// Probability normal = new Probability(DistributionType.NORMAL,
		// normalPara, normalPercentiles, normalX, normalProb);
		// return normal.inverse;
	}

	/**
	 * Computes a P-value of a value with the help of mean, standard deviation
	 * on the given distribution.
	 * 
	 * @param mean
	 *            of the distribution
	 * @param standardDeviation
	 *            of the distribution
	 * @param value
	 *            of the returned p-value
	 * @return p-value on the parameters
	 */
	protected double pValue(double mean, double standardDeviation, double value) {

		double percentile;
		double pValue;

		if (this.method.equals("GOODMAN")
				|| this.method.equals("GOODMAN+BONFERRONI")) {
			if (this.method.equals("GOODMAN+BONFERRONI")) {
				percentile = Math.sqrt(this.qchisq());
				double z = this
						.qnorm(this.alpha / (2 * this.numberOfContrasts));
				if (z <= percentile) {
					return pnorm(mean, standardDeviation, value);
				}
			}

			pValue = pchisq(mean, standardDeviation, value);

		} else if (this.method.equals("BONFERRONI")
				|| this.method.equals("HOLM-BONFERRONI")) {
			if (mean >= 0) {
				pValue = pnorm(-mean, standardDeviation, value);
			} else {
				pValue = pnorm(mean, standardDeviation, value);
			}
		} else {
			System.err.println("Unknown method! Using \"GOODMAN\"");
			pValue = pchisq(mean, standardDeviation, value);
		}

		if (Double.toString(pValue).equals("null")) {
			pValue = 1.0;
		}
		return pValue;

	}

	/**
	 * Computes the cumulative probability of the normal distribution. This
	 * method is needed to construct a p-value.
	 * 
	 * @param mean
	 *            of the normal distribution
	 * @param standardDeviation
	 *            of the normal distribution
	 * @param value
	 *            for the cumulative probability
	 * @return cumulative probability of the parameters using a normal
	 *         distribution
	 */
	private static double pnorm(double mean, double standardDeviation,
			double value) {
		if (standardDeviation > 0.0) {
			NormalDistribution normal = new NormalDistribution(mean, Math.pow(
					standardDeviation, 2.0));
			return (1.0 - normal.cumulative(value));
		} else {
			return 0.0;
		}
	}

	/**
	 * Computes the cumulative probability of the chi-square distribution. This
	 * method is needed to construct a p-value.
	 * 
	 * @param mean
	 *            of the chi-square distribution
	 * @param standardDeviation
	 *            of the chi-square distribution
	 * @param value
	 *            for the cumulative probability
	 * @return cumulative probability of the parameters using a chi-square
	 *         distribution
	 */
	private double pchisq(double mean, double standardDeviation, double value) {

		ChiSqrDistribution chisquare = new ChiSqrDistribution(
				this.degreeOfFreedom);

		return 1.0 - chisquare.cumulative(Math.pow((value - mean)
				/ standardDeviation, 2.0));
	}

	/**
	 * Getter of the method variable.
	 * 
	 * @return The used method
	 */
	protected String getMethod() {
		return this.method;
	}

	/**
	 * @param confidenceIntervals
	 */
	protected void holmBonferroniCorrection(
			Map<String, Map<String, Map<String, ConfidenceInterval>>> confidenceIntervals) {
		ArrayList<Double> pvalues = new ArrayList<Double>();
		for (Map<String, Map<String, ConfidenceInterval>> population2 : confidenceIntervals
				.values()) {
			for (Map<String, ConfidenceInterval> classes : population2.values()) {
				for (ConfidenceInterval ci : classes.values()) {
					pvalues.add(ci.pValue(0.0));
				}
			}
		}
		Collections.sort(pvalues);
		double numberOfHypothesis = this.numberOfContrasts;
		this.pValueBorder = this.alpha / numberOfHypothesis;
		while (pvalues.get(0) < this.pValueBorder) {
			pvalues.remove(0);
			numberOfHypothesis -= 0.5; // p-values are present two times p1-p2
										// and p2-p1
			this.pValueBorder = this.alpha / (Math.round(numberOfHypothesis));
		}
	}

	/**
	 * Getter of p-value border variable
	 * 
	 * @return The p-value border of the holm-bonferromi correction.
	 */
	protected double getPValueBorder() {
		return this.pValueBorder;
	}

	/**
	 * getter of comparisons
	 * 
	 * @return Number of comparisons of one class (n over 2)
	 */
	protected int getComparisons() {
		return this.comparisons;
	}

}
